import datetime as dt
from typing import Any, Dict, List, Optional, cast

import numpy as np
import pandas as pd


def repack_frame(
    df: pd.DataFrame,
    remap: Optional[Dict[str, str]] = None,
    dtypes: Optional[Dict[str, Any]] = {},
) -> pd.DataFrame:
    """
    Convert the DataFrame's columns to the most compact types possible.
    Rename columns if necessary during the repacking. The column renames
    work even if the column is part of the index.

    Args:
        remap: remap column names
        dtypes: dictionary of fixed dtypes to use
    """
    if df.index.names != [None]:
        raise ValueError("repacking is lost for index columns")

    remap = remap or {}
    dtypes = dtypes or {}

    # unwind the primary key
    if len(df.index.names) == 1 and not df.index.names[0]:
        primary_key = []
    else:
        primary_key = cast(List[str], df.index.names)
        df.reset_index(inplace=True)

    # repack each column into the best dtype we can give it
    df = pd.concat(
        [repack_series(df.loc[:, col]) if col not in dtypes else df[col] for col in df.columns],
        axis=1,
    )

    # use given dtypes
    if dtypes:
        df = df.astype(dtypes)

    # remap all column names, including those in the primary key
    for from_, to_ in remap.items():
        if from_ in df.columns:
            df.rename(columns={from_: to_}, inplace=True)
    primary_key = [remap.get(k, k) for k in primary_key]

    for col in df.columns:
        if df[col].dtype == "object":
            raise ValueError(f"Column {col} is still object. Consider converting it to str.")

    # set the primary key back again
    if primary_key:
        df.set_index(primary_key, inplace=True)

    return df


def repack_series(s: pd.Series) -> pd.Series:
    dtype_name = s.dtype.name.replace("[pyarrow]", "").replace("[pyarrow_numpy]", "").lower()

    if dtype_name in ("int64", "uint64"):
        return shrink_integer(s.astype("Int64"))

    if dtype_name in ("object", "str", "string", "float64"):
        for strategy in [to_int, to_float, to_category]:
            try:
                return strategy(s)
            except (ValueError, TypeError, OverflowError):
                continue

    return s


def _to_float(s: pd.Series) -> pd.Series:
    """Convert series to Float64. Replace numpy NaNs with NA. This can
    happen when original series is an object and contains 'nan' string."""
    r = s.astype("Float64")
    if s.dtype == "object":
        r = r.mask(np.isnan(r), pd.NA)
    return r


def to_int(s: pd.Series) -> pd.Series:
    # values could be integers or strings
    s = _to_float(s)
    v = s.astype("Int64")

    # casting to float converts strings to floats, that doesn't work with float64[pyarrow]
    if not series_eq(v, s):
        raise ValueError()

    # it's an integer, now pack it smaller
    return shrink_integer(v)


def shrink_integer(s: pd.Series) -> pd.Series:
    """
    Take an Int64 series and make it as small as possible.
    """
    assert s.dtype == "Int64"

    if s.isnull().all():
        # shrink all NaNs to Int8
        return s.astype("Int8")
    else:
        if s.min() < 0:
            series = ["Int32", "Int16", "Int8"]
        else:
            series = ["UInt32", "UInt16", "UInt8"]

    for dtype in series:
        v = s.astype(dtype)
        if not (v == s).all():
            break

        s = v

    return s


def to_float(s: pd.Series) -> pd.Series:
    return shrink_float(_to_float(s))


def shrink_float(s: pd.Series) -> pd.Series:
    """
    Take a Float64 series and make it as small as possible.
    """
    assert s.dtype.name.replace("[pyarrow]", "") in ("float64", "Float64", "double"), s.dtype

    options = ["Float32", "Float64"]
    for dtype in options:
        v = s.astype(dtype)

        if series_eq(s, v):
            return v

    raise ValueError()


def to_category(s: pd.Series) -> pd.Series:
    types = set(s.dropna().apply(type).unique())

    if types.difference({str, np.str_, dt.datetime, dt.date, type(None)}):
        raise ValueError()

    return s.astype("category")


def series_eq(lhs: pd.Series, rhs: pd.Series, rtol: float = 1e-5, atol: float = 1e-8) -> bool:
    """
    Check that series are equal, but unlike normal floating point checks where
    NaN != NaN, we want missing or null values to be reported as equal to each
    other.
    """
    # NOTE: this could be speeded up with numpy methods or smarter comparison,
    # but it's not bottleneck at the moment
    if len(lhs) != len(rhs):
        return False

    return np.allclose(lhs, rhs, rtol=rtol, atol=atol, equal_nan=True)


def _safe_dtype(dtype: Any) -> str:
    """Determine the appropriate dtype string based on pandas dtype."""
    if pd.api.types.is_integer_dtype(dtype):
        return "Int64"
    elif pd.api.types.is_float_dtype(dtype):
        return "Float64"
    elif pd.api.types.is_bool_dtype(dtype):
        return "boolean"
    elif isinstance(dtype, pd.CategoricalDtype):
        return "string[pyarrow]"
    elif dtype == "object":
        return "string[pyarrow]"
    else:
        return dtype


def to_safe_types(t: pd.DataFrame) -> pd.DataFrame:
    """Convert numeric columns to Float64 and Int64 and categorical
    columns to string[pyarrow]."""
    t = t.astype({col: _safe_dtype(t[col].dtype) for col in t.columns})

    if isinstance(t.index, pd.MultiIndex):
        t.index = t.index.set_levels([level.astype(_safe_dtype(level.dtype)) for level in t.index.levels])
    else:
        t.index = t.index.astype(_safe_dtype(t.index.dtype))

    return t



import datetime as dt
from typing import Any

import numpy as np
import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

from owid import repack


def test_repack_non_object_columns():
    df = pd.DataFrame({"myint": [1, 2, 3], "myfloat": [1.0, 2.2, 3.0], "mycat": ["a", "b", "c"]})
    df["mycat"] = df["mycat"].astype("category")

    df2 = df.copy()
    df2 = repack.repack_frame(df2, {})

    assert df2.myint.dtype.name == "UInt8"
    assert df2.myfloat.dtype.name == "Float32"
    assert_frame_equal(df, df2, check_dtype=False)


def test_repack_object_columns():
    df = pd.DataFrame(
        {
            "myint": [1, 2, None, 3],
            "myfloat": [1.2, 2.0, 3.0, None],
            "mycat": ["a", None, "b", "c"],
            "myintstr": [1, 2, 3, "4"],
            "nans1": [1, 2, 3, pd.NA],
            "nans2": [1, 2.1, 3, np.nan],
        },
        dtype="object",
    )

    df_repack = df.copy()

    df_repack = repack.repack_frame(df_repack)
    assert df_repack.myint.dtype.name == "UInt8"
    assert df_repack.myfloat.dtype.name == "Float32"
    assert df_repack.mycat.dtype.name == "category"
    assert df_repack.myintstr.dtype.name == "UInt8"
    assert df_repack.nans1.dtype.name == "UInt8"
    assert df_repack.nans2.dtype.name == "Float32"


def test_repack_frame_with_index():
    df = pd.DataFrame(
        {
            "myint": [1, 2, None, 3],
            "myfloat": [1.2, 2.0, 3.0, None],
            "mycat": ["a", None, "b", "c"],
        },
        dtype="object",
    )
    df.set_index(["myint", "mycat"], inplace=True)

    with pytest.raises(ValueError):
        repack.repack_frame(df)


def test_repack_integer_strings():
    s = pd.Series(["1", "2", "3", None])
    v = repack.repack_series(s)
    assert v.dtype.name == "UInt8"


def test_repack_float_strings():
    s = pd.Series(["10", "22.2", "30"])
    v = repack.repack_series(s)
    assert v.dtype.name == "Float32"


def test_repack_uint64():
    s = pd.Series([10, 20], dtype="uint64")
    v = repack.repack_series(s)
    assert v.dtype.name == "UInt8"


def test_repack_int8_boundaries():
    s = pd.Series([0, -1])
    info: Any = np.iinfo(np.int8)

    # check the lower boundary
    s[0] = info.min
    assert repack.repack_series(s).dtype.name == "Int8"
    s[0] -= 1
    assert repack.repack_series(s).dtype.name == "Int16"

    # check the upper boundary
    s[0] = info.max
    assert repack.repack_series(s).dtype.name == "Int8"
    s[0] += 1
    assert repack.repack_series(s).dtype.name == "Int16"


def test_repack_int16_boundaries():
    s = pd.Series([0, -1])
    info: Any = np.iinfo(np.int16)

    # check the lower boundary
    s[0] = info.min
    assert repack.repack_series(s).dtype.name == "Int16"
    s[0] -= 1
    assert repack.repack_series(s).dtype.name == "Int32"

    # check the upper boundary
    s[0] = info.max
    assert repack.repack_series(s).dtype.name == "Int16"
    s[0] += 1
    assert repack.repack_series(s).dtype.name == "Int32"


def test_repack_int32_boundaries():
    s = pd.Series([0, -1])
    info: Any = np.iinfo(np.int32)

    # check the lower boundary
    s[0] = info.min
    assert repack.repack_series(s).dtype.name == "Int32"
    s[0] -= 1
    assert repack.repack_series(s).dtype.name == "Int64"

    # check the upper boundary
    s[0] = info.max
    assert repack.repack_series(s).dtype.name == "Int32"
    s[0] += 1
    assert repack.repack_series(s).dtype.name == "Int64"


def test_repack_uint_boundaries():
    s = pd.Series([0])
    # uint8
    info: Any = np.iinfo(np.uint8)
    s[0] = info.max
    assert repack.repack_series(s).dtypes.name == "UInt8"

    s[0] += 1
    assert repack.repack_series(s).dtypes.name == "UInt16"

    # uint16
    info2: Any = np.iinfo(np.uint16)
    s[0] = info2.max
    assert repack.repack_series(s).dtypes.name == "UInt16"

    s[0] += 1
    assert repack.repack_series(s).dtypes.name == "UInt32"

    # uint32
    info3: Any = np.iinfo(np.uint32)
    s[0] = info3.max
    assert repack.repack_series(s).dtypes.name == "UInt32"

    # we don't bother using uint64, we just use int64
    s[0] += 1
    assert repack.repack_series(s).dtypes.name == "Int64"


def test_repack_int():
    s = pd.Series([1, 2, None, 3]).astype("object")
    v = repack.repack_series(s)
    assert v.dtype == "UInt8"


def test_repack_int_no_null():
    s = pd.Series([1, 2, 3]).astype("object")
    v = repack.repack_series(s)
    assert v.dtype == "UInt8"


def test_repack_float_to_int():
    s = pd.Series([1, 2, None, 3])
    assert s.dtype == "float64"
    v = repack.repack_series(s)
    assert v.dtype == "UInt8"


def test_repack_float_object_to_float32():
    s = pd.Series([1, 2, None, 3.3], dtype="object")

    v = repack.repack_series(s)
    assert v.dtype == "Float32"


def test_repack_object_with_nan_string():
    s = pd.Series([1, 2, "nan"], dtype="object")
    v = repack.repack_series(s)
    assert v.dtype == "UInt8"
    assert v.isnull().sum() == 1

    s = pd.Series([1, 2.2, "nan"], dtype="object")
    v = repack.repack_series(s)
    assert v.dtype == "Float32"
    assert v.isnull().sum() == 1


def test_repack_category():
    s = pd.Series(["a", "b", "c", None])
    assert s.dtype == np.object_ or s.dtype == "str"

    v = repack.repack_series(s)
    assert v.dtype == "category"


def test_shrink_integers_uint8():
    s = pd.Series([1, 2, 3], dtype="Int64")
    v = repack.shrink_integer(s)
    assert v.dtype.name == "UInt8"


def test_shrink_integers_int8():
    s = pd.Series([1, 2, 3, -3], dtype="Int64")
    v = repack.shrink_integer(s)
    assert v.dtype.name == "Int8"


def test_repack_frame_keep_dtypes():
    df = pd.DataFrame({"myint": [1, 2, 3], "myfloat": [1.0, 2.2, 3.0]})

    df2 = df.copy()
    df2 = repack.repack_frame(df2, dtypes={"myint": float})

    assert df2.myint.dtype.name == "float64"
    assert df2.myfloat.dtype.name == "Float32"


def test_repack_int64_all_nans():
    s = pd.Series([np.nan, np.nan, np.nan], dtype="Int64")
    v = repack.repack_series(s)
    assert v.dtype.name == "Int8"


def test_repack_float64_all_nans():
    s = pd.Series([np.nan, np.nan, np.nan], dtype="float64")
    v = repack.repack_series(s)
    assert v.dtype.name == "Int8"


def test_series_eq():
    a = pd.Series([1, np.nan], dtype="float64")
    b = pd.Series([2, np.nan], dtype="float64")
    assert not repack.series_eq(a, b)

    a = pd.Series([1, np.nan], dtype="float64")
    b = pd.Series([1, np.nan], dtype="float64")
    assert repack.series_eq(a, b)

    a = pd.Series([1, np.nan], dtype="float64")
    b = pd.Series([1, np.nan], dtype="float64").astype("Float64")
    assert repack.series_eq(a, b)


def test_repack_object_np_str():
    s = pd.Series(["a", np.str_("b")], dtype=object)
    v = repack.repack_series(s)
    assert v.dtype.name == "category"


def test_repack_with_inf():
    s = pd.Series([0, np.inf], dtype=object)
    v = repack.repack_series(s)
    assert v.dtype.name == "Float32"


def test_repack_with_datetime():
    s = pd.Series([dt.datetime.today(), dt.date.today()], dtype=object)
    v = repack.repack_series(s)
    assert v.dtype.name == "category"


def test_repack_string_type():
    s = pd.Series(["a", "b", "c"]).astype("string")
    assert s.dtype == "string"

    v = repack.repack_series(s)
    assert v.dtype == "category"


def test_to_safe_types():
    # Create a DataFrame with various dtypes
    df = pd.DataFrame(
        {
            "int_col": [1, 2, 3],
            "float_col": [1.1, 2.2, 3.3],
            "cat_col": pd.Categorical(["a", "b", "c"]),
            "object_col": ["x", "y", "z"],
        }
    )

    # Set an index with integer dtype
    df.set_index("int_col", inplace=True)

    # Apply the to_safe_types function
    df_safe = repack.to_safe_types(df)

    # Check that the dtypes have been converted appropriately
    assert df_safe.index.dtype == "Int64"
    assert df_safe["float_col"].dtype == "Float64"
    assert df_safe["cat_col"].dtype == "string[pyarrow]"
    assert df_safe["object_col"].dtype == "string[pyarrow]"


def test_to_safe_types_multiindex():
    # Create a DataFrame with MultiIndex
    df = pd.DataFrame(
        {
            "int_col": [1, 2, 3],
            "cat_col": pd.Categorical(["a", "b", "c"]),
            "float_col": [1.1, 2.2, 3.3],
        }
    )
    df.set_index(["int_col", "cat_col"], inplace=True)

    # Apply the to_safe_types function
    df_safe = repack.to_safe_types(df)

    # Check index levels
    assert df_safe.index.levels[0].dtype == "Int64"  # type: ignore
    assert df_safe.index.levels[1].dtype == "string[pyarrow]"  # type: ignore
    # Check column dtype
    assert df_safe["float_col"].dtype == "Float64"


def test_to_safe_types_with_nan():
    # Create a DataFrame with NaN values
    df = pd.DataFrame(
        {
            "int_col": [1, 2, 3],
            "int_with_nan": [1, np.nan, 3],
            "nullable_int_with_nan": [1, np.nan, 3],
            "float_col": [1.1, np.nan, 3.3],
            "cat_col": pd.Categorical(["a", None, "c"]),
        }
    )
    df.set_index("float_col", inplace=True)
    df["nullable_int_with_nan"] = df["nullable_int_with_nan"].astype("Int32")

    # Apply the to_safe_types function
    df_safe = repack.to_safe_types(df)

    # Check that NaN values are handled correctly
    assert df_safe.index.dtype == "Float64"
    assert df_safe["int_col"].dtype == "Int64"
    # NOTE: ints with nans end up as floats, but if they are nullable ints they remain as Int64 (which is our case
    # since we store everything as nullable types)
    assert df_safe["int_with_nan"].dtype == "Float64"
    assert df_safe["nullable_int_with_nan"].dtype == "Int64"
    assert df_safe["cat_col"].dtype == "string[pyarrow]"

    # Ensure that the NA value in 'cat_col' remains pd.NA and not the string "NA"
    assert pd.isna(df_safe["cat_col"].iloc[1])
    assert df_safe["cat_col"].iloc[1] is pd.NA
